/*
 * Clock and power management. This exposes the combined functionality
 * provided by the jz4730. The power management functionality could be exposed
 * using a separate driver.
 *
 * Copyright (C) 2017, 2018, 2020, 2021, 2023,
 *               2024 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/devices/hw_mmio_register_block.h>
#include "cpm-jz4730.h"



enum Regs : unsigned
{
  Clock_control       = 0x000,  // CFCR (CPCCR in JZ4740)
  Pll_control         = 0x010,  // PLCR1 (CPPCR in JZ4740)
  Exclk_oscillator    = 0x01c,  // OCR
  Clock_gate          = 0x020,  // MSCR (CLKGR in JZ4740)
  Sleep_control       = 0x024,  // SCR
  Lcd_divider         = 0x060,  // CFCR2
};



// Encoded value tables.

static const uint32_t dv[] = {1, 2, 3, 4, 6, 8, 12, 16, 24, 32, 1, 1, 1, 1, 1, 1};
static const uint32_t plldv[] = {1, 2, 2, 4};



// Register field definitions.

static Field Clock_source_ssi          (Clock_control, 1, 31),                // SCS
             Clock_source_lcd          (Clock_control, 1, 30),                // LCS
             Clock_source_i2s          (Clock_control, 1, 29),                // I2CS
             Clock_source_usb          (Clock_control, 1, 28),                // UCS
             Clock_source_msc          (Clock_control, 1, 24),                // MCS
             Clock_source_rtc          (Exclk_oscillator, 1, 8),              // O2SE

             Clock_change_enable       (Clock_control, 1, 20),                // UPE

             Clock_divider_cpu         (Clock_control, 0x0f, 0, dv),          // CFR
             Clock_divider_hclock      (Clock_control, 0x0f, 4, dv),          // HFR
             Clock_divider_pclock      (Clock_control, 0x0f, 8, dv),          // PFR
             Clock_divider_lcd         (Clock_control, 0x0f, 12, false, 1),   // LFR
             Clock_divider_mclock      (Clock_control, 0x0f, 16, dv),         // MFR
             Clock_divider_usb         (Clock_control, 0x07, 25, false, 1),   // UFR
             Clock_divider_lcd_pixel   (Lcd_divider, 0x1ff, 0),               // PXFR

             Clock_gate_uprt           (Clock_gate, 1, 25, true),
             Clock_gate_udc            (Clock_gate, 1, 24, true),
             Clock_gate_cim            (Clock_gate, 1, 23, true),
             Clock_gate_kbc            (Clock_gate, 1, 22, true),
             Clock_gate_emac           (Clock_gate, 1, 21, true),
             Clock_gate_uart3          (Clock_gate, 1, 20, true),
             Clock_gate_aic_bitclk     (Clock_gate, 1, 18, true),
             Clock_gate_scc            (Clock_gate, 1, 14, true),
             Clock_gate_msc            (Clock_gate, 1, 13, true),
             Clock_gate_ssi            (Clock_gate, 1, 12, true),
             Clock_gate_pwm1           (Clock_gate, 1, 11, true),
             Clock_gate_pwm0           (Clock_gate, 1, 10, true),
             Clock_gate_aic_pclk       (Clock_gate, 1, 9, true),
             Clock_gate_i2c            (Clock_gate, 1, 8, true),
             Clock_gate_lcd            (Clock_gate, 1, 7, true),
             Clock_gate_uhc            (Clock_gate, 1, 6, true),
             Clock_gate_dmac           (Clock_gate, 1, 5, true),
             Clock_gate_timer          (Clock_gate, 1, 3, true),
             Clock_gate_uart2          (Clock_gate, 1, 2, true),
             Clock_gate_uart1          (Clock_gate, 1, 1, true),
             Clock_gate_uart0          (Clock_gate, 1, 0, true),
             Clock_gate_sdram          (Clock_control, 1, 22),

             Pll_enable                (Pll_control, 1, 8),                   // PLLEN
             Pll_stable                (Pll_control, 1, 10),                  // PLLS
             Pll_bypass                (Pll_control, 1, 9),                   // PLLBP

             // Multiplier and input divider yield 2-based values.

             Pll_multiplier            (Pll_control, 0x1ff, 23, false, 2),    // PLLFD
             Pll_input_division        (Pll_control, 0x1f, 18, false, 2),     // PLLRD
             Pll_output_division       (Pll_control, 0x03, 16, plldv);        // PLLOD



// Multiplexer instances.

#define Clocks(...) ((enum Clock_identifiers []) {__VA_ARGS__})
#define Specific(CLOCK) ((enum Clock_identifiers) (CLOCK))

static Mux mux_external   (Clock_external),

           // Clocks being propagated to others.

           mux_core       (Clock_pll),
           mux_mclock     (Clock_mclock),
           mux_hclock     (Clock_hclock0),
           mux_pclock     (Clock_pclock),

           // Clock selection.

           mux_i2s        (2, Clocks(Clock_pll, Specific(Clock_pll_div_2))),
           mux_lcd_pixel  (2, Clocks(Clock_lcd_pixel0, Specific(Clock_lcd_pixel_in))),
           mux_msc        (2, Clocks(Specific(Clock_msc_16MHz), Specific(Clock_msc_24MHz))),
           mux_rtc        (2, Clocks(Clock_external_div, Clock_rtc_external)),
           mux_ssi        (2, Clocks(Clock_external, Clock_usb_phy)),
           mux_usb        (2, Clocks(Specific(Clock_pll_usb), Specific(Clock_usb_phy_48MHz)));



// Clock instances.
// NOTE: Leaving the LCD pixel clock as a null clock.

static Clock_null       clock_none,
                        clock_lcd_pixel_in;

static Clock_passive    clock_external(3686400),
                        clock_mac(25000000),
                        clock_msc_16MHz(16000000),
                        clock_msc_24MHz(24000000),
                        clock_rtc_external(32768),
                        clock_usb_phy_48MHz(48000000);

// Note the use of extra parentheses due to the annoying C++ "most vexing parse"
// problem. See: https://en.wikipedia.org/wiki/Most_vexing_parse

static Clock
                        clock_aic_bitclk((Source(mux_i2s)), Control(Clock_gate_aic_bitclk)),
                        clock_aic_pclk((Source(mux_i2s)), Control(Clock_gate_aic_pclk)),
                        clock_cim((Source(mux_hclock)), Control(Clock_gate_cim)),
                        clock_dma((Source(mux_external)), Control(Clock_gate_dmac)),
                        clock_emac((Source(mux_mclock)), Control(Clock_gate_emac)),
                        clock_i2c0((Source(mux_external)), Control(Clock_gate_i2c)),
                        clock_kbc((Source(mux_pclock)), Control(Clock_gate_kbc)),
                        clock_msc0((Source(mux_msc)), Control(Clock_gate_msc)),
                        clock_pwm0((Source(mux_external)), Control(Clock_gate_pwm0)),
                        clock_pwm1((Source(mux_external)), Control(Clock_gate_pwm1)),
                        clock_rtc((Source(mux_rtc))),
                        clock_scc((Source(mux_external)), Control(Clock_gate_scc)),
                        clock_sdram((Source(mux_mclock)), Control(Clock_gate_sdram)),
                        clock_ssi0((Source(mux_ssi)), Control(Clock_gate_ssi)),

                        // Timer propagates pclock but OST may select external
                        // or rtc instead.

                        clock_timer((Source(mux_pclock)), Control(Clock_gate_timer)),
                        clock_uart0((Source(mux_external)), Control(Clock_gate_uart0)),
                        clock_uart1((Source(mux_external)), Control(Clock_gate_uart1)),
                        clock_uart2((Source(mux_external)), Control(Clock_gate_uart2)),
                        clock_uart3((Source(mux_external)), Control(Clock_gate_uart3)),
                        clock_udc((Source(mux_usb)), Control(Clock_gate_udc)),
                        clock_uhc((Source(mux_usb)), Control(Clock_gate_uhc)),
                        clock_uprt((Source(mux_external)), Control(Clock_gate_uprt)),
                        clock_usb_phy((Source(mux_usb)));

static Clock_divided
                        clock_cpu(Source(mux_core),
                                  Control(Field::undefined, Clock_change_enable),
                                  Divider(Clock_divider_cpu)),
                        clock_hclock(Source(mux_core), 
                                     Control(Field::undefined, Clock_change_enable),
                                     Divider(Clock_divider_hclock)),
                        clock_lcd(Source(mux_core),
                                  Control(Clock_gate_lcd, Clock_change_enable),
                                  Divider(Clock_divider_lcd)),
                        clock_lcd_pixel(Source(mux_core),
                                        Control(Field::undefined, Clock_change_enable),
                                        Divider(Clock_divider_lcd_pixel)),
                        clock_mclock(Source(mux_core), 
                                     Control(Field::undefined, Clock_change_enable),
                                     Divider(Clock_divider_mclock)),
                        clock_pclock(Source(mux_core), 
                                     Control(Field::undefined, Clock_change_enable),
                                     Divider(Clock_divider_pclock)),
                        clock_pll_usb(Source(mux_core),
                                      Control(Field::undefined, Clock_change_enable),
                                      Divider(Clock_divider_usb));

static Clock_divided_fixed
                        clock_external_div((Source(mux_external)), (Divider_fixed(128))),
                        clock_pll_div_2((Source(mux_core)), (Divider_fixed(2)));

const double jz4730_pll_intermediate_min = 100000000,
             jz4730_pll_intermediate_max = 500000000;

static Pll   clock_pll(Source(mux_external),
                       Control_pll(Pll_enable, Pll_stable, Pll_bypass),
                       Divider_pll(Pll_multiplier, Pll_input_division,
                                   Pll_output_division,
                                   jz4730_pll_intermediate_min, jz4730_pll_intermediate_max));



// Clock register.

static Clock_base *clocks[Clock_jz4730_identifier_count] = {
  &clock_none,

  &clock_none,            // Clock_aic
  &clock_aic_bitclk,
  &clock_aic_pclk,
  &clock_none,            // Clock_can0
  &clock_none,            // Clock_can1
  &clock_none,            // Clock_cdbus
  &clock_cim,
  &clock_cpu,
  &clock_sdram,           // Clock_ddr
  &clock_dma,
  &clock_emac,
  &clock_external,
  &clock_external_div,
  &clock_hclock,          // Clock_hclock0
  &clock_none,            // Clock_hclock2
  &clock_none,            // Clock_hclock2_pclock
  &clock_none,            // Clock_hdmi
  &clock_i2c0,
  &clock_none,            // Clock_i2c1
  &clock_none,            // Clock_i2c2
  &clock_none,            // Clock_i2c3
  &clock_none,            // Clock_i2c4
  &clock_none,            // Clock_i2s0
  &clock_none,            // Clock_i2s0_rx
  &clock_none,            // Clock_i2s0_tx
  &clock_none,            // Clock_i2s1
  &clock_none,            // Clock_i2s1_rx
  &clock_none,            // Clock_i2s1_tx
  &clock_none,            // Clock_kbc
  &clock_none,            // Clock_l2cache
  &clock_lcd,
  &clock_lcd_pixel,       // Clock_lcd_pixel0
  &clock_none,            // Clock_lcd_pixel1
  &clock_mac,
  &clock_none,            // Clock_main
  &clock_none,            // Clock_mipi_csi
  &clock_none,            // Clock_msc
  &clock_msc0,
  &clock_none,            // Clock_msc1
  &clock_none,            // Clock_msc2
  &clock_none,            // Clock_nemc
  &clock_none,            // Clock_otg0
  &clock_none,            // Clock_otg1
  &clock_pclock,
  &clock_none,            // Clock_pcm
  &clock_pll,
  &clock_none,            // Clock_pll_A
  &clock_none,            // Clock_pll_E
  &clock_none,            // Clock_pll_M
  &clock_none,            // Clock_pll_V
  &clock_pwm0,
  &clock_pwm1,
  &clock_rtc,
  &clock_rtc_external,
  &clock_none,            // Clock_sadc
  &clock_scc,
  &clock_none,            // Clock_sfc
  &clock_none,            // Clock_ssi
  &clock_ssi0,
  &clock_none,            // Clock_ssi1
  &clock_none,            // Clock_ssi2
  &clock_none,            // Clock_timer
  &clock_uart0,
  &clock_uart1,
  &clock_uart2,
  &clock_uart3,
  &clock_none,            // Clock_uart4
  &clock_none,            // Clock_udc
  &clock_uhc,
  &clock_uprt,
  &clock_usb_phy,
  &clock_none,            // Clock_vpu

  /* JZ4730-specific clocks. */

  &clock_lcd_pixel_in,
  &clock_msc_16MHz,
  &clock_msc_24MHz,
  &clock_pll_div_2,
  &clock_pll_usb,
  &clock_usb_phy_48MHz,
};


// Peripheral abstraction.

Cpm_jz4730_chip::Cpm_jz4730_chip(l4_addr_t addr)
: Cpm_chip(addr, clocks)
{
}

Cpm_chip *
jz4730_cpm_chip(l4_addr_t cpm_base)
{
  return new Cpm_jz4730_chip(cpm_base);
}



// C language interface functions.

void *
jz4730_cpm_init(l4_addr_t cpm_base)
{
  return (void *) jz4730_cpm_chip(cpm_base);
}

const char *
jz4730_cpm_clock_type(void *cpm, enum Clock_identifiers clock)
{
  return static_cast<Cpm_chip *>(cpm)->clock_type(clock);
}

int
jz4730_cpm_have_clock(void *cpm, enum Clock_identifiers clock)
{
  return static_cast<Cpm_chip *>(cpm)->have_clock(clock);
}

void
jz4730_cpm_start_clock(void *cpm, enum Clock_identifiers clock)
{
  static_cast<Cpm_chip *>(cpm)->start_clock(clock);
}

void
jz4730_cpm_stop_clock(void *cpm, enum Clock_identifiers clock)
{
  static_cast<Cpm_chip *>(cpm)->stop_clock(clock);
}

int
jz4730_cpm_get_parameters(void *cpm, enum Clock_identifiers clock, uint32_t parameters[])
{ 
  return static_cast<Cpm_chip *>(cpm)->get_parameters(clock, parameters);
}

int
jz4730_cpm_set_parameters(void *cpm, enum Clock_identifiers clock, int num_parameters, uint32_t parameters[])
{ 
  return static_cast<Cpm_chip *>(cpm)->set_parameters(clock, num_parameters, parameters);
}

uint8_t
jz4730_cpm_get_source(void *cpm, enum Clock_identifiers clock)
{ 
  return static_cast<Cpm_chip *>(cpm)->get_source(clock);
}

void
jz4730_cpm_set_source(void *cpm, enum Clock_identifiers clock, uint8_t source)
{ 
  static_cast<Cpm_chip *>(cpm)->set_source(clock, source);
}

enum Clock_identifiers
jz4730_cpm_get_source_clock(void *cpm, enum Clock_identifiers clock)
{ 
  return static_cast<Cpm_chip *>(cpm)->get_source_clock(clock);
}

void
jz4730_cpm_set_source_clock(void *cpm, enum Clock_identifiers clock, enum Clock_identifiers source)
{ 
  static_cast<Cpm_chip *>(cpm)->set_source_clock(clock, source);
}

uint64_t
jz4730_cpm_get_source_frequency(void *cpm, enum Clock_identifiers clock)
{ 
  return static_cast<Cpm_chip *>(cpm)->get_source_frequency(clock);
}

uint64_t
jz4730_cpm_get_frequency(void *cpm, enum Clock_identifiers clock)
{ 
  return static_cast<Cpm_chip *>(cpm)->get_frequency(clock);
}

int
jz4730_cpm_set_frequency(void *cpm, enum Clock_identifiers clock, uint64_t frequency)
{
  return static_cast<Cpm_chip *>(cpm)->set_frequency(clock, frequency);
}
